;;; nhexl-mode.el --- Minor mode to edit files via hex-dump format

;; Copyright (C) 2010  Stefan Monnier

;; Author: Stefan Monnier <monnier@iro.umontreal.ca>
;; Keywords: data

;; This program is free software; you can redistribute it and/or modify
;; it under the terms of the GNU General Public License as published by
;; the Free Software Foundation, either version 3 of the License, or
;; (at your option) any later version.

;; This program is distributed in the hope that it will be useful,
;; but WITHOUT ANY WARRANTY; without even the implied warranty of
;; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
;; GNU General Public License for more details.

;; You should have received a copy of the GNU General Public License
;; along with this program.  If not, see <http://www.gnu.org/licenses/>.

;;; Commentary:

;; 

;;; Code:

(eval-when-compile (require 'cl))
(require 'hexl)                         ;For faces.

(defgroup nhexl nil
  "Edit a file in a hex dump format."
  :group 'data)

(defvar nhexl-line-width 16
  "Number of bytes per line.")

(defvar nhexl-display-table
  (let ((dt (make-display-table)))
    ;; (aset dt ?\n [?␊])
    (aset dt ?\t [?␉])
    dt))

(defvar nhexl-saved-vars nil)
(make-variable-buffer-local 'nhexl-saved-vars)
(defvar nhexl-point nil)
(make-variable-buffer-local 'nhexl-point)

;;;###autoload
(define-minor-mode nhexl-mode
  "Minor mode to edit files via hex-dump format"
  :lighter "NHexl"
  (if (not nhexl-mode)
      (progn
        (dolist (varl nhexl-saved-vars)
          (set (make-local-variable (car varl)) (cdr varl)))
        (kill-local-variable 'nhexl-saved-vars)
        (jit-lock-unregister 'nhexl-jit)
        (remove-hook 'after-change-functions 'nhexl-change-function 'local)
        (remove-hook 'post-command-hook 'nhexl-post-command 'local)
        ;; FIXME: This will conflict with any other use of `display'.
        (with-silent-modifications
          (put-text-property (point-min) (point-max) 'display nil))
        (remove-overlays (point-min) (point-max) 'nhexl t))
    (unless (local-variable-p 'nhexl-saved-vars)
      (dolist (var '(buffer-display-table buffer-invisibility-spec
                     overwrite-mode))
        (push (cons var (symbol-value var)) nhexl-saved-vars)))
    (setq nhexl-point (point))
    (binary-overwrite-mode 1)
    (setq buffer-invisibility-spec ())
    (set (make-local-variable 'buffer-display-table) nhexl-display-table)
    (jit-lock-register 'nhexl-jit)
    (add-hook 'change-major-mode-hook (lambda () (nhexl-mode -1)) nil 'local)
    (add-hook 'post-command-hook 'nhexl-post-command nil 'local)
    (add-hook 'after-change-functions 'nhexl-change-function nil 'local)))

(defun nhexl-change-function (beg end len)
  ;; Jit-lock already takes care of refreshing the changed area, so we
  ;; only have to make sure the tail's addresses are refreshed when
  ;; text is inserted/removed.
  (when (/= len (- end beg))
    (put-text-property beg (point-max) 'fontified nil)))

(defun nhexl-jit (from to)
  (let ((zero (save-restriction (widen) (point-min))))
    (setq from (+ zero (* (truncate (- from zero) nhexl-line-width)
                          nhexl-line-width)))
    (setq to (+ zero (* (ceiling (- to zero) nhexl-line-width)
                        nhexl-line-width)))
    (remove-overlays from (min to (point-max)) 'nhexl t)
    (save-excursion
      (goto-char from)
      (while (search-forward "\n" to t)
        (put-text-property (match-beginning 0) (match-end 0)
                           'display (copy-sequence "␊"))))
    (while (< from to)
      (let* ((next (+ from nhexl-line-width))
             (nextpos (min next (point-max)))
             (ol (make-overlay from next))
             (bufstr (buffer-substring from nextpos))
             (i -1)
             (str (concat (format (propertize "%08x:" 'face 'hexl-address-region)
                                  (- from zero))
                          (propertize " " 'face 'default
                                      'display '(space :align-to 12))
                          (mapconcat (lambda (c)
                                       (setq i (1+ i))
                                       ;; FIXME: In multibyte buffers,
                                       ;; do something clever about
                                       ;; non-ascii chars.
                                       (let ((s (format "%02x" c)))
                                         (when (= nhexl-point (+ from i))
                                           (put-text-property 0 (length s)
                                                              'face 'highlight
                                                              s))
                                         (if (zerop (mod i 2))
                                             s (concat s " "))))
                                     bufstr
                                     "")
                          (propertize "  " 'face 'default
                                      'display `(space :align-to
                                                       ,(+ 12 (/ (* nhexl-line-width 5) 2) 3))))))
                                      
        (if (> next nextpos)
            (setq str (concat str (make-string (+ (/ (1+ (- next nextpos)) 2)
                                                  (* (- next nextpos) 2))
                                               ?\s))))
        ;; (font-lock-append-text-property 0 (length str) 'face 'default str)
        (overlay-put ol 'nhexl t)
        ;; FIXME: This `face' setting messes up the display fairly
        ;; majorly (some text gets drawn more than once).
        (overlay-put ol'face 'hexl-ascii-region)
        (overlay-put ol 'before-string str)
        (overlay-put ol 'after-string (propertize "\n" 'face 'default))
        (setq from next)))))

(defun nhexl-post-command ()
  (when (/= (point) nhexl-point)
    (let ((zero (save-restriction (widen) (point-min)))
          (oldpoint nhexl-point))
      (setq nhexl-point (point))
      (with-silent-modifications
        (nhexl-jit (point) (1+ (point)))
        (if (/= (truncate (- (point) zero) nhexl-line-width)
                (truncate (- oldpoint zero) nhexl-line-width))
            (nhexl-jit oldpoint (1+ oldpoint)))))))
  

(provide 'nhexl-mode)
;;; nhexl-mode.el ends here
